# -*- coding: utf-8 -*-
""" Frontend related commands.

All frontend commands should be implemented through
``src/frontend/package.json``. This is just a proxy to allow easily running
them from the root project directory.
"""
from __future__ import absolute_import
from . import cli, click


@cli.command('fe')
@click.argument('cmd')
def fe(cmd):
    """ Run a predefined frontend command.

    The commands can be defined through the frontend.commands project
    configuration value. It should be a dict of commands mapped to the actual
    scripts ran inside the frontend.path.

    This command is mostly helpful if frontend code base is a subdirectory of
    the project. This is a very thin wrapper around javascript task runners,
    but it allows to run those commands from within any subdirectory of the
    project (as the root is defined by ``pelconf.py`` and not ``package.json``.

    If you only run frontend commands from within the frontend directory then
    this command is probably an unnecessary overhead.

    Sample Config::

        \b
        conf.init({
            'frontend': {
                'path': 'client/webui',
                'commands': {
                    'build': 'npm run build',
                    'watch': 'npm run watch'
                }
            }
        })

    Examples::

        \b
        $ peltak fe build       # Run frontend command named 'build'
        $ peltak fe watch       # Run frontend command named 'watch'
    """
    from .impl import fe

    fe.fe(cmd)


@cli.command('init-fe')
@click.argument('cmd')
@click.option(
    '--no-recreate', 'skip_if_exists',
    is_flag=True,
    help="Do nothing if node_modules already exists"
)
def init_fe(skip_if_exists):
    """ Initialize frontend.

    This is a little shortcut for ``npm init``. The main difference is that
    by default it will not do anything if ``node_modules`` directory is already
    present. Useful for CI runs when you have your node_modules cached and you
    know nothing changed but you still want to run it if the ``node_modules``
    directory is missing.

    Config Sample::

        \b
        conf.init({
            'frontend': {
                'path': 'client/webui'
            }
        })

    Examples::

        \b
        $ peltak init-fe                # Run 'npm-init'
        $ peltak init-fe --no-recreate  # Run only if node_modules is missing

    """
    from .impl import fe

    fe.init_fe(skip_if_exists)



# -*- coding: utf-8 -*-
"""
Django related management commands.

Those commands replace the usage of ``./manage.py`` (thus it's removed). Those
correspond 1 to 1 to their ``./manage.py`` counterparts. Those commands mainly
exists so that manage.py can be deleted (less top-level files in project dir).
"""
from __future__ import absolute_import
from . import cli, click


@cli.group('django')
def django_cli():
    """ Commands related to django """
    pass


@django_cli.command('devserver')
@click.option(
    '-p', '--port',
    type=int,
    default=8000,
    help="Port the server will run on"
)
@click.option(
    '-s', '--settings',
    type=str,
    help="Settings module to use. Defaults to DJANGO_SETTINGS conf variable."
)
def devserver(port=8000, settings=None):
    """ Run dev server. """
    _manage_cmd(['runserver', '0.0.0.0:{}'.format(port)], settings)


@django_cli.command('collectstatic')
def collectstatic():
    """ Collect all static files.

    This command won't ask for permission as manage.py does.

    Sample Config::

        \b
        conf.init({
            'SRC_DIR': './src'
            'DJANGO_SETTINGS': 'mypkg.settings',
        })

    Example::

        $ peltak django collectstatic

    """
    _manage_cmd(['collectstatic', '--no-input'])


@django_cli.command('mkmigrations')
@click.argument('app')
@click.argument('name')
def mkmigrations(app, name):
    """ Create a named migration for a given app.

    This will require the user to name every migration he creates thus improving
    the source code. Otherwise it's just a simple wrapper around ./manage.py.

    Sample Config::

        \b
        conf.init({
            'SRC_DIR': './src'
            'DJANGO_SETTINGS': 'mypkg.settings',
        })

    Example::

        $ peltak django mkmigrations mypkg.app add_new_fields_to_mytable

    """
    _manage_cmd(['makemigrations', '-n', name] + app.split(' '))


@django_cli.command('migrate')
def migrate():
    """ Apply pending migrations.

    Sample Config::

        \b
        conf.init({
            'SRC_DIR': './src'
            'DJANGO_SETTINGS': 'mypkg.settings',
        })

    Example::

        $ peltak django migrate

    """
    _manage_cmd(['migrate'])


@django_cli.command('createsuperuser')
def createsuperuser():
    """ Create super user (probably needed for the first user).

    Sample Config::

        \b
        conf.init({
            'SRC_DIR': './src'
            'DJANGO_SETTINGS': 'mypkg.settings',
        })

    Example::

        $ peltak django createsuperuser

    """
    _manage_cmd(['createsuperuser'])


@django_cli.command('shell')
def shell():
    """ Start django shell

    Sample Config::

        \b
        conf.init({
            'SRC_DIR': './src'
            'DJANGO_SETTINGS': 'mypkg.settings',
        })

    Example::

        $ peltak django shell

    """
    _manage_cmd(['shell'])


def _manage_cmd(cmd, settings=None):
    """ Run django ./manage.py command manually.

    This function eliminates the need for having ``manage.py`` (reduces file
    clutter).
    """
    import sys
    from os import environ
    from peltak.core import conf

    sys.path.insert(0, conf.get('SRC_DIR'))

    settings = settings or conf.get('DJANGO_SETTINGS', None)
    environ.setdefault("DJANGO_SETTINGS_MODULE", settings)

    from django.core.management import execute_from_command_line

    args = sys.argv[0:-1] + cmd

    execute_from_command_line(args)



# -*- coding: utf-8 -*-
""" Docker commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys

# 3rd party imports
import click

# local imports
from peltak.core import shell
from peltak.core import log
from peltak.core import conf
from peltak.core import docker


def build_images():
    """ Build all docker images for the project. """
    registry = conf.get('docker.registry')
    docker_images = conf.get('docker.images', [])

    for image in docker_images:
        docker.build_image(registry, image)


def push_images():
    """ Push all project docker images to a remote registry. """
    registry = conf.get('docker.registry')
    docker_images = conf.get('docker.images', [])

    if registry is None:
        log.err("You must define docker.registry conf variable to push images")
        sys.exit(-1)

    for image in docker_images:
        docker.push_image(registry, image)


def docker_list(registry_pass):
    """ List docker images stored in the remote registry. """
    registry = conf.get('docker.registry', None)

    if registry is None:
        log.err("You must define docker.registry conf variable to list images")
        sys.exit(-1)

    registry_user = conf.get('docker.registry_user', None)

    if registry_user is None:
        registry_user = click.prompt("Username")

    rc = docker.RegistryClient(registry, registry_user, registry_pass)
    images = {x: rc.list_tags(x) for x in rc.list_images()}

    shell.cprint("<32>Images in <34>{} <32>registry:", registry)
    for image, tags in images.items():
        shell.cprint('  <92>{}', image)
        for tag in tags:
            shell.cprint('      <90>{}:<35>{}', image, tag)



# -*- coding: utf-8 -*-
""" Commands for building and pushing the app image.

This is mainly for web apps that are deployed with docker.
"""
from __future__ import absolute_import
from . import cli, click


@cli.group('docker')
def docker_cli():
    """ Commands for building the project docker images. """
    pass


@docker_cli.command('build')
def build_images():
    """ Build and tag a docker image for the project.

    It requires a docker.images conf variable to be set and contain the
    definitions for all the images built by the project. If docker.registry is
    also set, then the resulting images will a have name of
    ``<registry>/<image>``. This will create two tags for each image:
    ``<name>:<version>`` and ``<name>:latest`` where <name> is the image name
    as specified in docker.images entry and <version> is the current project
    version (as shown by ``peltak version``).

    Config example::

        \b
        conf.init({
            'docker': {
                'registry': 'registry.mydomain.com',
                'images': [
                    {'name': 'myapp'}
                ]
            }
        })

    Example::

        \b
        $ peltak docker build

    """
    from .impl import docker

    docker.build_images()


@docker_cli.command('push')
def push_images():
    """ Push project docker images to the registry.

    This command requires both docker.images and docker.registry conf variables
    to be set. This will push the images built by ``peltak docker build`` to the
    specified registry. For the details of how the names and tags are generated,
    see ``peltak docker build --help``.

    Config example::

        \b
        conf.init({
            'docker': {
                'registry': 'registry.mydomain.com',
                'images': [
                    {'name': 'myapp'}
                ]
            }
        })

    Example::

        \b
        $ peltak docker push

    """
    from .impl import docker

    docker.push_images()


@docker_cli.command('list')
@click.option(
    '-p', '--password', 'registry_pass',
    type=str,
    prompt='Password',
    hide_input=True,
    help='Registry password'
)
def docker_list(registry_pass):
    """ List docker images and their tags on the remote registry.

    This command requires the docker.registry conf variable to be set. You can
    also hard-code the username with the docker.registry_user conf variable. You
    will be asked for the password every time though for security reasons.

    Config example::

        \b
        conf.init({
            'docker': {
                'registry': 'registry.mydomain.com',
                'registry_user': 'myuser'
            }
        })

    Example::

        \b
        $ peltak docker list            # List images in the docker registry
        $ peltak docker list -p mypass  # List registry images, use give pw

    """
    from .impl import docker

    docker.docker_list(registry_pass)



# -*- coding: utf-8 -*-
""" Commands implementation.

Kept in a separate package to reduce the number of imports and file sizes for
the commands module. This way the commands module contain only code used by
click to generate the completion and everything else is imported when used.
"""
from __future__ import absolute_import, unicode_literals



# -*- coding: utf-8 -*-
""" Helpers for related docker commands. """
from __future__ import absolute_import, unicode_literals

# local imports
from . import conf
from . import log
from . import shell
from . import versioning


class RegistryClient(object):
    """ Helper class for talking to docker registry. """
    def __init__(self, registry, username, password):
        self.user = username
        self.pw = password
        self.auth = (username, password)
        self.registry_url = 'https://' + registry

    def get(self, *args, **kw):
        """ Proxy over requests.get

        Do not include requests globally so it's not required to build and
        push images. Otherwise, the project needs to install requests even
        when not using registry in any way.
        """
        import requests
        return requests.get(*args, **kw)

    def list_images(self):
        """ List images stored in the registry.

        :return List[str]:
            List of image names.
        """
        r = self.get(self.registry_url + '/v2/_catalog', auth=self.auth)
        return r.json()['repositories']

    def list_tags(self, image_name):
        """ List all tags for the given image stored in the registry.

        :param str image_name:
            The name of the image to query. The image must be present on the
            registry for this call to return any values.
        :return List[str]:
            List of tags for that image.
        """
        tags_url = self.registry_url + '/v2/{}/tags/list'

        r = self.get(tags_url.format(image_name), auth=self.auth)
        data = r.json()

        if 'tags' in data:
            return reversed(sorted(data['tags']))

        return []


def build_image(registry, image):
    """ Build docker image.

    :param str registry:
        The name of the registry this image belongs to. If not given, the
        resulting image will have a name without the registry.
    :param Dict image:
        The dict containing the information about the built image. This is the
        same dictionary as defined in DOCKER_IMAGES variable.
    """
    if ':' in image['name']:
        name, tag = image['name'].split(':', 1)
    else:
        name, tag = image['name'], None

    values = {
        'registry': '' if registry is None else registry + '/',
        'image': name,
        'tag': tag,
        'version': versioning.current(),
    }

    if tag is not None:
        args = [
            '-t {registry}{image}:{tag}'.format(**values),
        ]
    else:
        args = [
            '-t {registry}{image}'.format(**values),
            '-t {registry}{image}:{version}'.format(**values),
        ]

    if 'file' in image:
        args.append('-f {}'.format(conf.proj_path(image['file'])))

    with conf.within_proj_dir(image.get('path', '.')):
        log.info("Building <33>{registry}<35>/{image}".format(**values))
        shell.run('docker build {args} .'.format(args=' '.join(args)))


def push_image(registry, image):
    """ Push the given image to selected repository.

    :param str registry:
        The name of the registry we're pushing to. This is the address of the
        repository without the protocol specification (no http(s)://)
    :param Dict image:
        The dict containing the information about the image. This is the same
        dictionary as defined in DOCKER_IMAGES variable.
    """
    values = {
        'registry': registry,
        'image': image['name'],
    }

    log.info("Pushing <33>{registry}<35>/{image}".format(**values))
    shell.run('docker push {registry}/{image}'.format(**values))



# -*- coding: utf-8 -*-
""" Frontend commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from os.path import exists, join

# local imports
from peltak.core import conf
from peltak.core import log
from peltak.core import shell


def fe(cmd):
    """ Run a frontend command.

    The commands are defined in the project config.

    :param str cmd:
        Frontend command to run.
    """
    frontend_cmds = conf.get('frontend.commands', {
        'build': 'npm run build',
        'start': 'npm start',
        'watch': 'npm run watch',
        'test': 'npm test',
        'lint': 'npm lint',
        'check': 'npm run lint && npm run test'
    })

    if cmd in frontend_cmds:
        _fe_cmd(frontend_cmds[cmd])
    else:
        log.err("No {} in frontend.commands".format(cmd))


def init_fe(skip_if_exists):
    """ Initialize frontend.

    This is a little shortcut for ``npm init``. The main difference is that
    by default it will not do anything if ``node_modules`` directory is already
    present. Useful for CI runs when you have your node_modules cached and you
    know nothing changed but you still want to run it if the ``node_modules``
    directory is m

    :param skip_if_exists:
    :return:
    """
    frontend_path = conf.get_path('frontend.path', None)
    initialized = exists(join(frontend_path, 'node_modules'))

    if not initialized or not skip_if_exists:
        _fe_cmd('npm install')


def _fe_cmd(cmd):
    frontend_path = conf.get_path('frontend.path', None)

    if frontend_path is not None:
        with conf.within_proj_dir(frontend_path):
            shell.run(cmd)
    else:
        log.err("No frontend.path defined in the config")



# -*- coding: utf-8 -*-
""" AppEngine commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
from fnmatch import fnmatch
from os.path import exists, join, normpath

# local imports
from peltak.core import conf
from peltak.core import fs
from peltak.core import git
from peltak.core import log
from peltak.core import shell
from peltak.core import versioning
from peltak.core import util


@util.mark_experimental
def devserver(port, admin_port, clear):
    """ Run devserver.

    :param int port:
        Port on which the app will be served.
    :param int admin_port:
        Port on which the admin interface is served.
    :param bool clear:
        If set to **True**, clear the datastore on startup.
    """
    admin_port = admin_port or (port + 1)

    args = [
        '--port={}'.format(port),
        '--admin_port={}'.format(admin_port)
    ]

    if clear:
        args += ['--clear_datastore=yes']

    with conf.within_proj_dir():
        shell.run('dev_appserver.py . {args}'.format(args=' '.join(args)))


@util.mark_experimental
def deploy(pretend, promote, deploy_all):
    """ Deploy the app to AppEngine.

    :param bool pretend:
        If set to **True**, do not actually deploy anything but show the deploy
        command that would be used.
    :param bool promote:
        If set to **True** promote the current remote app version to the one
        that's being deployed.
    :param bool deploy_all:
    :return:
    """
    gae_projects = conf.get('appengine.projects')
    branch = git.branch_details()
    gae_proj = next(
        (e for e in gae_projects if fnmatch(branch.name, e['branch'])),
        None
    )
    args = []

    if gae_proj is None:
        log.err("Can't find an GAE project setup for branch <35>{}", branch)
        sys.exit(1)

    if promote:
        args += ['--promote']
    else:
        args += ['--no-promote']

    if branch.type == 'feature':
        app_version = '{ver}-{title}'.format(
            ver=versioning.current().replace('.', '-'),
            title=branch.title.replace('_', '-')
        )
    else:
        app_version = versioning.current().replace('.', '-')

    args += [
        '--version {}'.format(app_version),
        '--project {}'.format(gae_proj['name']),
    ]

    deployables = [gae_proj['config']]

    if deploy_all:
        deployables += conf.get('appengine.deployables', [
            'cron.yaml',
            'index.yaml',
            'queue.yaml',
        ])

    with conf.within_proj_dir():
        cmd = 'gcloud app deploy {args} {deployables}'.format(
            deployables=fs.surround_paths_with_quotes(deployables),
            args=' '.join(args)
        )

        if pretend:
            log.info("Would deploy version <35>{ver} <32>to <35>{proj}".format(
                ver=app_version,
                proj=gae_proj['name']
            ))
            shell.cprint('<90>{}', cmd)

        if not pretend:
            log.info("Deploying version <35>{ver} <32>to <35>{proj}".format(
                ver=app_version,
                proj=gae_proj['name']
            ))
            shell.run(cmd)


@util.mark_experimental
def setup_ci():
    """ Setup AppEngine SDK on CircleCI """
    gcloud_path = shell.run('which gcloud', capture=True).stdout.strip()
    sdk_path = normpath(join(gcloud_path, '../../platform/google_appengine'))
    gcloud_cmd = gcloud_path + ' --quiet'

    if not exists(sdk_path):
        log.info("Installing AppEngine SDK")
        shell.run('sudo {} components install app-engine-python'.format(
            gcloud_cmd
        ))
    else:
        # Only initialise once. To reinitialise, just build without cache.
        log.info("AppEngine SDK already initialised")

    log.info("Using service account authentication")
    shell.run('{} auth activate-service-account --key-file {}'.format(
        gcloud_cmd,
        conf.proj_path('ops/client_secret.json')
    ))

    # shell.run('{} config set project {}'.format(gcloud_cmd, project))


def _find_appengine_sdk():
    gcloud_path = shell.run('which gcloud', capture=True).stdout.strip()
    return normpath(join(gcloud_path, '../../platform/google_appengine'))



# -*- coding: utf-8 -*-
""" Shell related functions """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import re
import subprocess
import sys
from collections import namedtuple


ExecResult = namedtuple(
    'ExecResult',
    'command return_code stdout stderr succeeded failed'
)


is_tty = sys.stdout.isatty()


def fmt(msg, *args, **kw):
    """ Generate shell color opcodes from a pretty coloring syntax. """
    global is_tty

    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    opcode_subst = '\x1b[\\1m' if is_tty else ''
    return re.sub(r'<(\d{1,2})>', opcode_subst, msg)


def cprint(msg, *args, **kw):
    """ Print colored message to stdout. """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    print(fmt('{}<0>'.format(msg)))


def run(cmd, capture=False, shell=True, env=None, exit_on_error=None):
    """ Run a shell command.

    :param Union[str, unicode] cmd:
        The shell command to execute.
    :param bool shell:
        Same as in `subprocess.Popen()`.
    :param bool capture:
        If set to True, it will capture the standard input/error instead of
        just piping it to the caller stdout/stderr.
    :param dict env:
        The subprocess environment variables.
    :param bool exit_on_error:
        If set to *True* (default), on failure it will call `sys.exit()` with
        the return code for the executed command.
    :return ExecResult:
        Return instance of ``ExecResult``. The result contains the return code
        and output (if capture was set to *True*).
    """
    options = {
        'bufsize': 1,       # line buffered
        'shell': shell
    }

    if exit_on_error is None:
        exit_on_error = not capture

    if capture:
        options.update({
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
        })

    if env is not None:
        options['env'] = dict(os.environ)
        options['env'].update(env)

    p = subprocess.Popen(cmd, **options)
    stdout, stderr = p.communicate()

    try:
        if stdout is not None:
            stdout = stdout.decode('utf-8')

        if stderr is not None:
            stderr = stderr.decode('utf-8')
    except AttributeError:
        # 'str' has no attribute 'decode'
        pass

    if exit_on_error and p.returncode != 0:
        sys.exit(p.returncode)

    return ExecResult(
        cmd,
        p.returncode,
        stdout,
        stderr,
        p.returncode == 0,
        p.returncode != 0
    )



# -*- coding: utf-8 -*-
""" Git helpers. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
from collections import namedtuple

# 3rd party imports
from six import string_types

# local imports
from . import conf
from . import shell


Author = namedtuple('Author', 'name email')
BranchDetails = namedtuple('BranchDetails', 'type title name')


def current_branch():
    """ Return the name of the currently checked out git branch. """
    cmd = 'git symbolic-ref --short HEAD'
    return shell.run(cmd, capture=True).stdout.strip()


def commit_author(sha1=''):
    """ Return the author of the given commit.

    :param str sha1:
        The sha1 of the commit to query. If not given, it will return the sha1
        for the current commit.
    :return Author:
        A named tuple ``(name, email)`` with the commit author details.
    """
    with conf.within_proj_dir():
        cmd = 'git show -s --format="%an||%ae" {}'.format(sha1)
        result = shell.run(cmd, capture=True).stdout
        name, email = result.split('||')
        return Author(name, email)


def untracked():
    """ Return a list of untracked files in the project repository.

    :return List[str]:
        The list of files not tracked by project git repo.
    """
    with conf.within_proj_dir():
        status = shell.run('git status --porcelain', capture=True).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status.startswith('?? '):
                results.append(file_status[3:].strip())

        return results


def staged():
    """ Return a list of project files staged for commit.

    :return List[str]:
        The list of project files staged for commit.
    """
    with conf.within_proj_dir():
        status = shell.run('git status --porcelain', capture=True).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status and file_status[0] in ('A', 'M'):
                results.append(file_status[3:].strip())

        return results


def ignore():
    """ Return a list of patterns in the project .gitignore

    :return List[str]:
        List of patterns set to be ignored by git.
    """

    def parse_line(line):   # pylint: disable=missing-docstring
        # Decode if necessary
        if not isinstance(line, string_types):
            line = line.decode('utf-8')

        # Strip comment
        line = line.split('#', 1)[0].strip()

        return line

    with conf.within_proj_dir():
        with open('.gitignore') as fp:
            return [parse_line(l) for l in fp.readlines() if l.strip()]


def branch_details():
    """

    :return BranchDetails:
    """
    branch_name = current_branch()
    if '/' in branch_name:
        branch_type, branch_title = branch_name.rsplit('/', 1)
        return BranchDetails(branch_type, branch_title, branch_name)

    return BranchDetails(branch_name, None, branch_name)



# -*- coding: utf-8 -*-
""" Documentation commands implementation. """
from __future__ import absolute_import, unicode_literals, print_function

# stdlib imports
import os.path
import shutil
import sys

# local imports
from peltak.core import conf
from peltak.core import log
from peltak.core import shell


def docs(recreate, gen_index):
    """ Build the documentation for the project.

    :param bool recreate:
        If set to **True**, the build and output directories will be cleared
        prior to generating the docs.
    :param bool gen_index:
        If set to **True**, it will generate top-level index file for the
        reference documentation.
    """
    build_dir = conf.get_path('build_dir', '.build')
    doc_src_path = conf.get_path('docs.path', 'docs')
    refdoc_paths = conf.get('docs.reference', [])

    docs_out_path = os.path.join(doc_src_path, 'html')
    docs_ref_path = os.path.join(doc_src_path, 'ref')
    docs_assets_path = os.path.join(doc_src_path, 'assets')
    docs_build_path = os.path.join(build_dir, 'docs')

    log.info('Ensuring assets directory <94>{}<32> exists', docs_assets_path)
    if not os.path.exists(docs_assets_path):
        os.makedirs(docs_assets_path)

    if recreate:
        for path in (docs_out_path, docs_build_path):
            if os.path.exists(path):
                log.info("<91>Deleting <94>{}".format(path))
                shutil.rmtree(path)

    if refdoc_paths:
        _gen_ref_docs(docs_ref_path, not gen_index)
    else:
        log.err('Not generating any reference documentation - '
                'No docs.reference specified in config')

    with conf.within_proj_dir(doc_src_path):
        log.info('Building docs with <35>sphinx')
        shell.run('sphinx-build -b html -d {build} {docs} {out}'.format(
            build=docs_build_path,
            docs=doc_src_path,
            out=docs_out_path,
        ))
        log.info('You can view the docs by browsing to <34>file://{}'.format(
            os.path.join(docs_out_path, 'index.html')
        ))


def _gen_ref_docs(ref_path, no_index=False):
    try:
        from refdoc import generate_docs as _generate_docs
    except ImportError as ex:
        msg = ("You need to install sphinx-refdoc if you want to generate "
               "code reference docs.")

        print(msg, file=sys.stderr)
        log.err("Exception: {}".format(ex))
        sys.exit(-1)

    refdoc_paths = conf.get('docs.reference', [])

    log.info('Removing previously generated reference documentation')
    if os.path.exists(ref_path):
        shutil.rmtree(ref_path)

    os.makedirs(ref_path)

    log.info('Generating reference documentation')
    args = {'out_dir': ref_path}

    if no_index:
        args['no_index'] = True

    pkg_paths = [conf.proj_path(p) for p in refdoc_paths]

    _generate_docs(pkg_paths, **args)



# -*- coding: utf-8 -*-
""" Time related utilities. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import time
import warnings
from functools import wraps


class timed_block(object):  # noqa
    """ Context manager to measure execution time for a give block of code.

    .. code-block:: python

        with timed_block() as t:
            time.sleep(1)

        print("The block took {}ms to execute".format(t.elapsed_ms)
    """
    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.elapsed = time.time() - self.t0
        self.elapsed_s = round(self.elapsed, 3)
        self.elapsed_ms = round(self.elapsed * 1000, 3)


def mark_experimental(fn):
    """ Mark function as experimental.

    :param Function fn:
        The command function to decorate.
    """
    @wraps(fn)
    def wrapper(*args, **kw):   # pylint: disable=missing-docstring
        warnings.warn("This command is has experimental status. The interface "
                      "is not yet stable and might change without notice and "
                      "with a minor version. Use at your own risk")
        return fn(*args, **kw)

    return wrapper


def mark_deprecated(replaced_by):
    """ Mark command as deprecated.

    :param str replaced_by:
        The command that deprecated this command and should be used instead.
    """
    def decorator(fn):   # pylint: disable=missing-docstring
        @wraps(fn)
        def wrapper(*args, **kw):   # pylint: disable=missing-docstring
            warnings.warn("This command is has been deprecated. Please use "
                          "{new} instead.".format(new=replaced_by))

            return fn(*args, **kw)

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
"""
Various commands that do not fit another category but there's not enough of them
to justify a separate module.
"""
from __future__ import absolute_import
from . import cli, click


@cli.command('docs')
@click.option(
    '--recreate',
    is_flag=True,
    help="Delete build and out directories before running."
)
@click.option(
    '--gen-index',
    is_flag=True,
    help="Generate main index for code reference"
)
def docs(recreate, gen_index):
    """ Build project documentation.

    This command will run sphinx-refdoc first to generate the reference
    documentation for the code base. Then it will run sphinx to generate the
    final docs. You can configure the directory that stores the docs source
    (index.rst, conf.py, etc.) using the DOC_SRC_PATH conf variable. In case you
    need it, the sphinx build directory is located in ``BUILD_DIR/docs``.

    The reference documentation will be generated for all directories listed
    under 'REFDOC_PATHS conf variable. By default it is empty so no reference
    docs are generated.

    Sample Config::

        \b
        conf.init({
            'build_dir': '.build',
            'docs': {
                'path': 'docs',
                'reference': [
                    'src/mypkg'
                ]
            }
        })

    Examples::

        \b
        $ peltak docs                           # Generate docs for the project
        $ peltak docs --no-index                # Skip main reference index
        $ peltak docs --recreate --no-index     # Build docs from clean slate

    """
    from .impl import docs

    docs.docs(recreate, gen_index)



# -*- coding: utf-8 -*-
"""
Testing commands
"""
from __future__ import absolute_import
from . import cli, click


@cli.command('test')
@click.argument('tests_type', metavar='<type>', type=str, default='default')
@click.option(
    '-v', '--verbose',
    count=True,
    help=("Be verbose. Can specify multiple times for more verbosity. This "
          "will also influence the verbosity of pytest output.")
)
@click.option(
    '--junit',
    is_flag=True,
    help="Write junit XML report to build_dir/test-results.xml."
)
@click.option(
    '--no-locals',
    is_flag=True,
    help="Disable pytest printing the values of local variables."
)
@click.option(
    '--no-coverage',
    is_flag=True,
    help="Disable code coverage generation."
)
@click.option(
    '--allow-empty',
    is_flag=True,
    help="Do not return non-zero code if the test suite is empty."
)
@click.option(
    '--plugins',
    type=str,
    default='',
    help=("Comma separated list of pytest plugins to activate. If the name "
          "of the plugins starts with '-' it will be disabled, otherwise it "
          "will be enabled.")
)
def test(
        tests_type, verbose, junit, no_locals, no_coverage, plugins, allow_empty
):
    """ Run tests against the current python version.

    This command uses pytest internally and is just a thing wrapper over it
    to simplify most common tasks. You can define multiple test types where
    a test type is a combination of paths to test and pytest marker definition
    for the test run. This allows to split tests with fine granularity and
    and run them easily. You need at least one test type to run the tests. The
    test type *default* will be used as default if no type is specified. If
    there is no configuration for *default* it will run the tests in the
    directory specified by SRC_PATH conf variable.

    src_dir is used as PYTHONPATH for the test runner. If you need to customize
    it, use that variable.

    The coverage report will be written to ``build_dir/coverage``.

    If your project is using django, you can use django_test_settings conf
    variable to specify which settings to use for the tests.

    Config Sample::

        \b
        conf.init({
            'coverage_cfg_path': 'ops/tools/coverage',
            'django_test_settings': 'mypkg.settings.test',
            'test': {
                'types': {
                    'default': {'paths': ['test']},
                    'no_django': {
                        'mark': 'not django',
                        'paths': ['test']
                    }
                }
            }
        })

    Examples::

        \b
        $ peltak test                   # Run tests using the default options
        $ peltak --no-sugar             # Disable pretty output
        $ peltak --junit                # Generate build_dir/test-results.xml
        $ peltak test --no-sugar -vv    # Be extra verbose

    """
    from .impl import test
    test.test(
        tests_type, verbose, junit, no_locals, no_coverage, plugins, allow_empty
    )



# -*- coding: utf-8 -*-
""" Commands related to Google AppEngine.

Only useful on appengine projects. If you're not using AppEngine, do not
import inside `pelconf.py`.
"""
from __future__ import absolute_import
from . import cli, click


@cli.group('appengine')
def appengine_cli():
    """ Google AppEngine related commands. """
    pass


@appengine_cli.command('deploy')
@click.option(
    '--pretend',
    is_flag=True,
    help=("Do not actually deploy to AppEngine.This will only collect all "
          "static files and compile i18n messages and tell you where the "
          "app would be deployed and what version would that be. Locally this "
          "is the same as just running the command but no changes will be done "
          "to the remote deployment.")
)
@click.option(
    '--promote',
    is_flag=True,
    help=("If specified, the currently deployed version will become the active "
          "one")
)
@click.option(
    '--all', 'deploy_all',
    is_flag=True,
    help=("If specified, the command will deploy all deployables, and not only "
          "the app file. If you don't specify it, the command will only deploy "
          "app.yaml (for the given env ofc). ")
)
def deploy(pretend, promote, deploy_all):
    """ Deploy the app to the target environment.

    The target environments can be configured using the ENVIRONMENTS conf
    variable. This will also collect all static files and compile translation
    messages
    """
    from .impl.appengine import deploy

    deploy(pretend, promote, deploy_all)


@appengine_cli.command()
@click.option('-p', '--port', type=int, default=8080)
@click.option('--admin-port', type=int, default=None)
@click.option('--clear', is_flag=True)
def devserver(port, admin_port, clear):
    """ Run devserver. """
    from .impl.appengine import devserver

    devserver(port, admin_port, clear)


@appengine_cli.command('setup-ci')
def setup_ci():
    """ Setup AppEngine SDK on CircleCI """
    from .impl.appengine import setup_ci

    setup_ci()



# -*- coding: utf-8 -*-
""" Miscellaneous commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
from os.path import exists, isdir
from shutil import rmtree

# 3rd party imports
import click

# local imports
from peltak.core import conf
from peltak.core import fs
from peltak.core import log
from peltak.core import shell


def clean(pretend, exclude):
    """ Remove all unnecessary files.

    :param bool pretend:
        If set to **True**, do not delete any files, just show what would be
        deleted.
    :param List[str] exclude:
        A list of path patterns to exclude from deletion.
    """
    clean_patterns = conf.get('clean_patterns', [
        '*__pycache__*',
        '*.py[cod]',
        '*.swp',
    ])

    files = fs.filtered_walk(conf.proj_path(), clean_patterns, exclude)

    for path in files:
        try:
            if not isdir(path):
                log.info('  <91>[file] <90>{}', path)
                if not pretend:
                    os.remove(path)
            else:
                log.info('  <91>[dir]  <90>{}', path)
                if not pretend:
                    rmtree(path)
        except OSError:
            log.info("<33>Failed to remove <90>{}", path)


def init():
    """ Create an empty pelconf.py from template """
    config_file = 'pelconf.py'
    prompt = "-- <35>{} <32>already exists. Wipe it?<0>".format(config_file)

    if exists(config_file) and not click.confirm(shell.fmt(prompt)):
        log.info("Canceled")
        return

    log.info('Writing <35>{}'.format(config_file))
    with open('pelconf.py', 'w') as fp:
        fp.write('''# -*- coding: utf-8 -*-
""" peltak configuration file.

See https://github.com/novopl/peltak for more information.
"""
from __future__ import absolute_import

# Configure the build
from peltak.core import conf

conf.init({
})

# Import all commands
#from peltak.commands import version
''')



# -*- coding: utf-8 -*-
""" Release commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
import os
from fnmatch import fnmatch
from os.path import join

# local imports
from peltak.core import shell
from peltak.core import conf
from peltak.core import git
from peltak.core import log
from peltak.core import versioning


def release(component, exact):
    """ Create a new release branch.

    :param str component:
        Version component to bump when creating the release. Can be *major*,
        *minor* or *patch*.
    :param str exact:
        The exact version to set for the release. Overrides the component
        argument. This allows to re-release a version if something went wrong
        with the release upload.
    """
    version_file = conf.get_path('version_file', 'VERSION')

    with conf.within_proj_dir():
        out = shell.run('git status --porcelain', capture=True).stdout
        lines = out.split(os.linesep)
        has_changes = any(
            not l.startswith('??') for l in lines if l.strip()
        )

    if has_changes:
        log.info("Cannot release: there are uncommitted changes")
        exit(1)

    old_ver, new_ver = versioning.bump(component, exact)

    log.info("Bumping package version")
    log.info("  old version: <35>{}".format(old_ver))
    log.info("  new version: <35>{}".format(new_ver))

    with conf.within_proj_dir():
        branch = 'release/' + new_ver

        log.info("Checking out new branch <35>{}", branch)
        shell.run('git checkout -b ' + branch)

        log.info("Creating commit for the release")

        shell.run('git add {ver_file} && git commit -m "{msg}"'.format(
            ver_file=version_file,
            msg="Releasing v{}".format(new_ver)
        ))


def tag_release():
    """ Tag the current commit with the release version. """
    release_ver = versioning.current()
    author = git.commit_author()

    with conf.within_proj_dir():
        log.info("Creating tag that marks the release")
        cmd = (
            'git -c "user.name={0.name}" -c "user.email={0.email}" '
            'tag -a "v{1}" -m "Mark v{1} release"'
        ).format(
            author,
            release_ver
        )
        shell.run(cmd)


def upload(target):
    """ Upload the release to a pypi server.

    TODO: Make sure the git directory is clean before allowing a release.

    :param str target:
        pypi target as defined in ~/.pypirc
    """
    log.info("Uploading to pypi server <33>{}".format(target))
    with conf.within_proj_dir():
        shell.run('python setup.py sdist register -r "{}"'.format(target))
        shell.run('python setup.py sdist upload -r "{}"'.format(target))


def gen_pypirc(username=None, password=None):
    """ Generate ~/.pypirc with the given credentials.

    Useful for CI builds. Can also get credentials through env variables
    ``PYPI_USER`` and ``PYPI_PASS``.

    :param str username:
        pypi username.
    :param str password:
        pypi password.
    """
    path = join(conf.getenv('HOME'), '.pypirc')
    username = username or conf.getenv('PYPI_USER', None)
    password = password or conf.getenv('PYPI_PASS', None)

    if username is None or password is None:
        log.err("You must provide $PYPI_USER and $PYPI_PASS")
        sys.exit(1)

    log.info("Generating .pypirc config <94>{}".format(path))

    with open(path, 'w') as fp:
        fp.write('\n'.join((
            '[distutils]',
            'index-servers = ',
            '    pypi',
            '',
            '[pypi]',
            'repository: https://upload.pypi.org/legacy/',
            'username: {}'.format(username),
            'password: {}'.format(password),
            '',
        )))


def merged():
    """ Cleanup the release branch after it was remotely merged to master. """
    develop_branch = conf.get('git.devel_branch', 'develop')
    master_branch = conf.get('git.master_branch', 'master')
    protected_branches = conf.get(
        'git.protected_branches',
        (master_branch, develop_branch)
    )
    release_branch_pattern = conf.get('git.release_branch', 'release/*')
    branch = git.current_branch()

    if not fnmatch(branch, release_branch_pattern):
        log.err("You can only run `peltak release merged` on release branches. "
                "You can specify The release branch pattern with "
                "release_branch_pattern conf variable (defaults to release/*).")
        sys.exit(1)

    try:
        shell.run('git rev-parse --verify ' + branch)
    except IOError:
        log.err("Branch '{}' does not exist".format(branch))

    # Checkout develop and merge the release
    log.info("Checking out <33>{}".format(develop_branch))
    shell.run('git checkout ' + develop_branch)

    log.info("Merging <35>{} <32>into <33>{}".format(branch, develop_branch))
    shell.run('git merge ' + branch)

    log.info("Checking out <33>{}".format(master_branch))
    shell.run('git checkout ' + master_branch)

    log.info("Pulling latest changes")
    shell.run('git pull origin ' + master_branch)

    if branch not in protected_branches:
        log.info("Deleting branch <35>{}".format(branch))
        shell.run('git branch -d ' + branch)

    log.info("Pruning")
    shell.run('git fetch --prune origin')

    log.info("Checking out <33>{}<32> branch".format(develop_branch))
    shell.run('git checkout ' + develop_branch)



# -*- coding: utf-8 -*-
""" Code linting commands. """
from __future__ import absolute_import
from . import cli, click


@cli.command('lint')
@click.option(
    '-e', '--exclude',
    multiple=True,
    metavar='PATTERN',
    help=("Specify patterns to exclude from linting. For multiple patterns, "
          "use the --exclude option multiple times")
)
@click.option(
    '--skip-untracked',
    is_flag=True,
    help="Also include files not tracked by git."
)
@click.option(
    '--commit', 'commit_only',
    is_flag=True,
    help=("Only lint files staged for commit. Useful if you want to clean up "
          "a large code base one commit at a time.")
)
@click.option(
    '-p', '--pretend',
    is_flag=True,
    help=("Just print files that would be linted without running anything")
)
def lint(exclude, skip_untracked, commit_only, pretend):
    """ Run pep8 and pylint on all project files.

    You can configure the linting paths using the lint.paths config variable.
    This should be a list of paths that will be linted. If a path to a directory
    is given, all files in that directory and it's subdirectories will be
    used.

    The pep8 and pylint config paths are by default stored in ops/tools/pep8.ini
    and ops/tools/pylint.ini. You can customise those paths in your config with
    lint.pep8_cfg and lint.pylint_cfg variables.

    **Config Example**::

        \b
        conf.init({
            'lint': {
                'pylint_cfg': 'ops/tools/pylint.ini',
                'pep8_cfg': 'ops/tools/pep8.ini',
                'paths': [
                    'src/mypkg'
                ],
            }
        })

    **Examples**::

        \b
        $ peltak lint               # Run linter in default mode, skip untracked
        $ peltak lint --commit      # Lint only files staged for commit
        $ peltak lint --all         # Lint all files, including untracked.
        $ peltak lint --pretend     # Print the list of files to lint
        $ peltak lint -e "*.tox*"   # Don't lint files inside .tox directory

    """
    from .impl import lint

    lint.lint(exclude, skip_untracked, commit_only, pretend)



# -*- coding: utf-8 -*-
""" Helper commands for releasing to pypi. """
from __future__ import absolute_import
from . import cli, click


@cli.group('release', invoke_without_command=True)
@click.option(
    '-c', '--component',
    type=click.Choice(['major', 'minor', 'patch']),
    required=False,
    default='patch',
    help=("Which version component to bump with this release.")
)
@click.option(
    '--exact',
    type=str,
    help="Set the newly released version to be exactly as specified."
)
@click.pass_context
def release_cli(ctx, component, exact):
    """ Create a new release branch.

    It will bump the current version number and create a release branch called
    `release/<version>` with one new commit (the version bump).

    **Example Config**::

        \b
        conf.init({
            'version_file': 'src/mypkg/__init__.py'
        })

    **Examples**::

        \b
        $ peltak release --component=patch    # Make a new patch release
        $ peltak release -c minor             # Make a new minor release
        $ peltak release -c major             # Make a new major release
        $ peltak release                      # same as release -c patch
        $ peltak release tag                  # Tag current commit as release
        $ peltak release upload pypi          # Upload to pypi

    """
    if ctx.invoked_subcommand:
        return

    from .impl import release
    release.release(component, exact)


@release_cli.command('tag')
def tag_release():
    """ Tag the current commit with as the current version release.

    This should be the same commit as the one that's uploaded as the release
    (to pypi for example).

    **Example Config**::

        \b
        conf.init({
            'version_file': 'src/mypkg/__init__.py'
        })

    Examples::

        $ peltak release tag          # Tag the current commit as release

    """
    from .impl import release
    release.tag_release()


@release_cli.command()
@click.argument('target')
def upload(target):
    """ Upload to a given pypi target.

    Examples::

        \b
        $ peltak release upload pypi    # Upload the current release to pypi
        $ peltak release upload private # Upload to pypi server 'private'

    """
    from .impl import release
    release.upload(target)


@release_cli.command('gen-pypirc')
@click.argument('username', required=False)
@click.argument('password', required=False)
def gen_pypirc(username=None, password=None):
    """
    Generate .pypirc config with the given credentials.

    Example::

        $ peltak release gen-pypirc my_pypi_user my_pypi_pass

    """
    from .impl import release
    release.gen_pypirc(username, password)


@release_cli.command('merged')
def merged():
    """ Checkout the target branch, pull and delete the merged branch.

    This is to ease the repetitive cleanup of each merged branch.

    Example Config (those the are defaults)::

        \b
        conf.init({
            'main_branch': 'develop',
            'master_branch': 'master',
            'protected_branches': ['master', 'develop'],
            'release_branch_pattern: 'release/*'
        })

    Example::

        $ peltak release merged     # Must be ran on the relase branch

    """
    from .impl import release
    release.merged()



# -*- coding: utf-8 -*-
""" Commands for managing the managed project version. """
from __future__ import absolute_import
from . import cli, click


@cli.group('version', invoke_without_command=True)
@click.option('--porcelain', is_flag=True)
@click.pass_context
def version_cli(ctx, porcelain):
    """ Show project version. Has sub commands.

    For this command to work you must specify where the project version is
    stored. You can do that with version_file conf variable. peltak supports
    multiple ways to store the project version. Right now you can store it in a
    python file using built-in __version__ variable. You can use node.js
    package.json and keep the version there or you can just use a plain text
    file that just holds the raw project version. The appropriate storage is
    guessed based on the file type and name.

    Example configuration::

        conf.init({
            'version_file': 'src/mypackage/__init__.py'
        })

    Examples:

        \b
        $ peltak version                        # Pretty print current version
        $ peltak version --porcelain            # Print version as raw string
        $ peltak version bump patch             # Bump patch version component
        $ peltak version bump minor             # Bump minor version component
        $ peltak version bump major             # Bump major version component
        $ peltak version bump release           # same as version bump patch
        $ peltak version bump --exact=1.2.1     # Set project version to 1.2.1

    """
    if ctx.invoked_subcommand:
        return

    from peltak.core import log
    from peltak.core import versioning

    current = versioning.current()

    if porcelain:
        print(current)
    else:
        log.info("Version: <35>{}".format(current))


@version_cli.command('bump')
@click.argument(
    'component',
    type=click.Choice(['major', 'minor', 'patch']),
    required=False,
    default='patch'
)
@click.option('--exact', type=str)
def bump_version(component='patch', exact=None):
    """ Bump current project version without committing anything.

    No tags are created either.

    Examples:

        \b
        $ peltak version bump patch             # Bump patch version component
        $ peltak version bump minor             # Bump minor version component
        $ peltak version bump major             # Bump major version component
        $ peltak version bump release           # same as version bump patch
        $ peltak version bump --exact=1.2.1     # Set project version to 1.2.1

    """
    from peltak.core import log
    from peltak.core import versioning

    old_ver, new_ver = versioning.bump(component, exact)

    log.info("Bumping package version")
    log.info("  old version: <35>{}".format(old_ver))
    log.info("  new version: <35>{}".format(new_ver))



# -*- coding: utf-8 -*-
""" Test command implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
import os.path

# local imports
from peltak.core import conf
from peltak.core import log
from peltak.core import fs
from peltak.core import shell


def test(tests_type,
         verbose,
         junit,
         no_locals,
         no_coverage,
         plugins,
         allow_empty):
    """ Run tests using pytest

    :param str tests_type:
        Tests type to run. The types are defined in the project configuration.
    :param int verbose:
        Verbosity level (0-3)
    :param bool junit:
        If **True** it will save junit test report in build directory.
    :param bool no_locals:
        Do not show locals
    :param bool no_coverage:
        Do not generate coverage report
    :param List[str] plugins:
        List of pytest plugins to enable/disable. If the plugin name is prefixed
        with '-' sign the plugin will be disabled.
    :param bool allow_empty:
        If set to **False** it will return non-zero code if the test suite is
        empty.
    """
    log.info("Running <33>{} <32>tests".format(tests_type))

    build_dir = conf.get_path('build_dir', '.build')

    pytest_cfg_path = conf.get_path('test.pytest_cfg', 'ops/tools/pytest.ini')
    test_types = conf.get('test.types', {})

    coverage_out_path = os.path.join(build_dir, 'coverage')
    coverage_cfg_path = conf.get_path('coverage_cfg_path',
                                      'ops/tools/coverage.ini')

    django_settings = conf.get('django_settings', None)
    django_test_settings = conf.get('django_test_settings', None)

    src_dir = conf.get_path('src_dir')
    src_path = conf.get_path('src_path')
    plugins = plugins.split(',')
    args = []

    if not no_coverage:
        args += [
            '--cov-config={}'.format(coverage_cfg_path),
            '--cov={}'.format(src_path),
            '--cov-report=term',
            '--cov-report=html:{}'.format(coverage_out_path),
        ]

    if junit:
        args += ['--junitxml={}/test-results.xml'.format('.build')]

    if '-django' not in plugins:
        if django_test_settings is not None:
            args += ['--ds {}'.format(django_test_settings)]
        elif django_settings is not None:
            args += ['--ds {}'.format(django_settings)]

    if verbose >= 1:
        args += ['-' + 'v' * verbose]
    if verbose >= 2:
        args += ['--full-trace']

    if not no_locals:
        args += ['-l']

    if plugins:
        for plug_name in plugins:
            if not plug_name.strip():
                continue

            if plug_name.startswith('-'):
                args += ['-p no:{}'.format(plug_name[1:])]
            else:
                args += ['-p {}'.format(plug_name)]

    test_config = test_types.get(tests_type)
    if test_config is None:
        log.info("Test type configuration missing: <34>{}".format(src_path))

        for possible_path in (conf.proj_path('test'), conf.proj_path('tests')):
            if os.path.exists(possible_path):
                test_config = {'paths': [possible_path]}
                break
        else:
            log.err("No tests detected. Configure your TEST_TYPES.")
            sys.exit(-1)

    mark = test_config.get('mark')

    if mark:
        args += ['-m "{}"'.format(mark)]

    test_paths = test_config['paths'] or []
    test_paths = [conf.proj_path(p) for p in test_paths]
    result = shell.run(
        'pytest -c {conf} {args} {paths}'.format(
            conf=pytest_cfg_path,
            args=' '.join(args),
            paths=fs.surround_paths_with_quotes(test_paths)
        ),
        env={'PYTHONPATH': src_dir},
        exit_on_error=False
    )

    log.info("To see the HTML coverage report browse to <34>file://{}".format(
        conf.proj_path('.build/coverage/index.html')
    ))

    if result.failed and not (allow_empty is True and result.return_code == 5):
        sys.exit(result.return_code)
